@@ -27,14 +27,6 @@
 #include "internal/chafa-pixops.h"
 #include "internal/chafa-work-cell.h"
 
-/* Used for cell initialization. May be added up over multiple cells, so a
- * low multiple needs to fit in an integer. */
-#define SYMBOL_ERROR_MAX (G_MAXINT / 8)
-
-/* Max candidates to consider in pick_symbol_and_colors_fast(). This is also
- * limited by a similar constant in chafa-symbol-map.c */
-#define N_CANDIDATES_MAX 8
-
 typedef struct
 {
     ChafaPixel fg;
