@@ -63,6 +63,9 @@
 /* Size of buffer used for copying stdin to file */
 #define COPY_BUFFER_SIZE 8192
 
+/* A contiguous magic string can't be longer than this */
+#define MAGIC_LENGTH_MAX 1024
+
 struct FileMapping
 {
     gchar *path;
@@ -139,7 +142,6 @@ safe_read (gint fd, void *buf, gsize len)
 static gboolean
 safe_write (gint fd, gconstpointer buf, gsize len)
 {
-    gsize ntotal = 0;
     const guint8 *buffer = buf;
     gboolean success = FALSE;
 
@@ -170,7 +172,6 @@ safe_write (gint fd, gconstpointer buf, gsize len)
            /* Continue writing until permanent failure or entire buffer written */
            buffer += n_written;
            len -= (unsigned int) n_written;
-           ntotal += (unsigned int) n_written;
        }
     }
 
@@ -368,7 +369,7 @@ open_file (FileMapping *file_mapping, GError **error)
         if (fd < 0)
         {
             g_set_error (error, G_FILE_ERROR, g_file_error_from_errno (errno),
-                         strerror (errno));
+                         "%s", strerror (errno));
         }
     }
 
@@ -607,7 +608,9 @@ file_mapping_read (FileMapping *file_mapping, gpointer out, goffset ofs, gsize l
 gboolean
 file_mapping_has_magic (FileMapping *file_mapping, goffset ofs, gconstpointer data, gsize length)
 {
-    gchar *buf;
+    gchar buf [MAGIC_LENGTH_MAX];
+
+    g_assert (length <= MAGIC_LENGTH_MAX);
 
     if (!ensure_open_file (file_mapping))
         return FALSE;
@@ -628,8 +631,6 @@ file_mapping_has_magic (FileMapping *file_mapping, goffset ofs, gconstpointer da
     if (lseek (file_mapping->fd, ofs, SEEK_SET) != ofs)
         return FALSE;
 
-    buf = alloca (length);
-
     if (safe_read (file_mapping->fd, buf, length) != length)
         return FALSE;
 
