@@ -121,7 +121,7 @@ png_loader_get_is_animation (PngLoader *loader)
     return FALSE;
 }
 
-gconstpointer *
+gconstpointer
 png_loader_get_frame_data (PngLoader *loader, ChafaPixelType *pixel_type_out,
                            gint *width_out, gint *height_out, gint *rowstride_out)
 {
